@@ -1,6 +1,7 @@
 /*
  * cups - Python bindings for CUPS
- * Copyright (C) 2002, 2005, 2006, 2007  Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2002, 2005, 2006, 2007, 2008, 2009, 2010, 2011  Red Hat, Inc
+ * Author: Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,53 +26,94 @@
 #include "cupsmodule.h"
 
 #include <locale.h>
+#include <pthread.h>
+#include <wchar.h>
+#include <wctype.h>
 
 #include "cupsconnection.h"
 #include "cupsppd.h"
+#include "cupsipp.h"
 
-static PyObject *cups_password_callback = NULL;
+static pthread_key_t tls_key = -1;
+static pthread_once_t tls_key_once = PTHREAD_ONCE_INIT;
 
 //////////////////////
 // Worker functions //
 //////////////////////
 
-static int
-do_model_compare (const char *a, const char *b)
+static void
+destroy_TLS (void *value)
 {
-  const char *digits = "0123456789";
-  char quick_a, quick_b;
-  while ((quick_a = *a) != '\0' && (quick_b = *b) != '\0') {
-    int end_a = strspn (a, digits);
-    int end_b = strspn (b, digits);
+  struct TLS *tls = (struct TLS *) value;
+  Py_XDECREF (tls->cups_password_callback);
+
+#if HAVE_CUPS_1_4
+  Py_XDECREF (tls->cups_password_callback_context);
+#endif /* HAVE_CUPS_1_4 */
+
+  free (value);
+}
+
+static void
+init_TLS (void)
+{
+  pthread_key_create (&tls_key, destroy_TLS);
+}
+
+struct TLS *
+get_TLS (void)
+{
+  struct TLS *tls;
+  pthread_once (&tls_key_once, init_TLS);
+  tls = (struct TLS *) pthread_getspecific (tls_key);
+  if (tls == NULL)
+    {
+      tls = calloc (1, sizeof (struct TLS));
+      pthread_setspecific (tls_key, tls);
+    }
+
+  return tls;
+}
+
+static int
+do_model_compare (const wchar_t *a, const wchar_t *b)
+{
+  const wchar_t *digits = L"0123456789";
+  wchar_t quick_a, quick_b;
+  while ((quick_a = *a) != L'\0' && (quick_b = *b) != L'\0') {
+    int end_a = wcsspn (a, digits);
+    int end_b = wcsspn (b, digits);
     int min;
     int a_is_digit = 1;
     int cmp;
 
-    if (quick_a != quick_b && !isdigit (quick_a) && !isdigit (quick_b)) {
+    if (quick_a != quick_b && !iswdigit (quick_a) && !iswdigit (quick_b)) {
       if (quick_a < quick_b) return -1;
       else return 1;
     }
 
     if (!end_a) {
-      end_a = strcspn (a, digits);
+      end_a = wcscspn (a, digits);
       a_is_digit = 0;
     }
 
     if (!end_b) {
       if (a_is_digit)
 	return -1;
-      end_b = strcspn (b, digits);
+      end_b = wcscspn (b, digits);
     } else if (!a_is_digit)
       return 1;
 
     if (a_is_digit) {
-      int n_a = atoi (a), n_b = atoi (b);
+      unsigned long n_a, n_b;
+      n_a = wcstoul (a, NULL, 10);
+      n_b = wcstoul (b, NULL, 10);
       if (n_a < n_b) cmp = -1;
       else if (n_a == n_b) cmp = 0;
       else cmp = 1;
     } else {
       min = end_a < end_b ? end_a : end_b;
-      cmp = strncmp (a, b, min);
+      cmp = wcsncmp (a, b, min);
     }
 
     if (!cmp) {
@@ -86,8 +128,8 @@ do_model_compare (const char *a, const char *b)
     return cmp;
   }
 
-  if (quick_a == '\0') {
-    if (*b == '\0')
+  if (quick_a == L'\0') {
+    if (*b == L'\0')
       return 0;
 
     return -1;
@@ -96,20 +138,28 @@ do_model_compare (const char *a, const char *b)
   return 1;
 }
 
+#ifndef HAVE_CUPS_1_4
 static const char *
 do_password_callback (const char *prompt)
 {
+  struct TLS *tls = get_TLS ();
   static char *password;
 
   PyObject *args;
   PyObject *result;
   const char *pwval;
 
+  debugprintf ("-> do_password_callback\n");
+  Connection_end_allow_threads (tls->g_current_connection);
   args = Py_BuildValue ("(s)", prompt);
-  result = PyEval_CallObject (cups_password_callback, args);
+  result = PyEval_CallObject (tls->cups_password_callback, args);
   Py_DECREF (args);
   if (result == NULL)
+  {
+    debugprintf ("<- do_password_callback (empty string)\n");
+    Connection_begin_allow_threads (tls->g_current_connection);
     return "";
+  }
 
   if (password) {
     free (password);
@@ -120,10 +170,17 @@ do_password_callback (const char *prompt)
   password = strdup (pwval);
   Py_DECREF (result);
   if (!password)
+  {
+    debugprintf ("<- do_password_callback (empty string)\n");
+    Connection_begin_allow_threads (tls->g_current_connection);
     return "";
-  
+  }
+
+  Connection_begin_allow_threads (tls->g_current_connection);
+  debugprintf ("<- do_password_callback\n");
   return password;
 }
+#endif /* !HAVE_CUPS_1_4 */
 
 //////////////////////////
 // Module-level methods //
@@ -132,12 +189,62 @@ do_password_callback (const char *prompt)
 static PyObject *
 cups_modelSort (PyObject *self, PyObject *args)
 {
-  char *a, *b;
+  PyObject *Oa, *Ob, *a, *b;
+  int len_a, len_b;
+  size_t size_a, size_b;
+  wchar_t *wca, *wcb;
 
-  if (!PyArg_ParseTuple (args, "ss", &a, &b))
+  if (!PyArg_ParseTuple (args, "OO", &Oa, &Ob))
     return NULL;
 
-  return Py_BuildValue ("i", do_model_compare (a, b));
+  a = PyUnicode_FromObject (Oa);
+  b = PyUnicode_FromObject (Ob);
+  if (a == NULL || b == NULL || !PyUnicode_Check (a) || !PyUnicode_Check (b)) {
+    if (a) {
+      Py_DECREF (a);
+    }
+    if (b) {
+      Py_DECREF (b);
+    }
+
+    PyErr_SetString (PyExc_TypeError, "Unable to convert to Unicode");
+    return NULL;
+  }
+
+  len_a = 1 + PyUnicode_GetSize (a);
+  size_a = len_a * sizeof (wchar_t);
+  if ((size_a / sizeof (wchar_t)) != len_a) {
+    Py_DECREF (a);
+    Py_DECREF (b);
+    PyErr_SetString (PyExc_RuntimeError, "String too long");
+    return NULL;
+  }
+
+  len_b = 1 + PyUnicode_GetSize (b);
+  size_b = len_b * sizeof (wchar_t);
+  if ((size_b / sizeof (wchar_t)) != len_b) {
+    Py_DECREF (a);
+    Py_DECREF (b);
+    PyErr_SetString (PyExc_RuntimeError, "String too long");
+    return NULL;
+  }
+
+  wca = malloc (size_a);
+  wcb = malloc (size_b);
+  if (wca == NULL || wcb == NULL) {
+    Py_DECREF (a);
+    Py_DECREF (b);
+    free (wca);
+    free (wcb);
+    PyErr_SetString (PyExc_RuntimeError, "Insufficient memory");
+    return NULL;
+  }
+
+  PyUnicode_AsWideChar ((PyUnicodeObject *) a, wca, size_a);
+  PyUnicode_AsWideChar ((PyUnicodeObject *) b, wcb, size_b);
+  Py_DECREF (a);
+  Py_DECREF (b);
+  return Py_BuildValue ("i", do_model_compare (wca, wcb));
 }
 
 static PyObject *
@@ -218,6 +325,7 @@ cups_getEncryption (PyObject *self)
 static PyObject *
 cups_setPasswordCB (PyObject *self, PyObject *args)
 {
+  struct TLS *tls = get_TLS ();
   PyObject *cb;
 
   if (!PyArg_ParseTuple (args, "O:cups_setPasswordCB", &cb))
@@ -228,10 +336,81 @@ cups_setPasswordCB (PyObject *self, PyObject *args)
     return NULL;
   }
 
+  debugprintf ("-> cups_setPasswordCB\n");
+#ifdef HAVE_CUPS_1_4
+  Py_XDECREF (tls->cups_password_callback_context);
+  tls->cups_password_callback_context = NULL;
+#endif /* HAVE_CUPS_1_4 */
+
   Py_XINCREF (cb);
-  Py_XDECREF (cups_password_callback);
-  cups_password_callback = cb;
+  Py_XDECREF (tls->cups_password_callback);
+  tls->cups_password_callback = cb;
+
+#ifdef HAVE_CUPS_1_4
+  cupsSetPasswordCB2 (password_callback_oldstyle, NULL);
+#else
   cupsSetPasswordCB (do_password_callback);
+#endif
+
+  debugprintf ("<- cups_setPasswordCB\n");
+  Py_INCREF (Py_None);
+  return Py_None;
+}
+
+#ifdef HAVE_CUPS_1_4
+static PyObject *
+cups_setPasswordCB2 (PyObject *self, PyObject *args)
+{
+  struct TLS *tls = get_TLS ();
+  PyObject *cb;
+  PyObject *cb_context = NULL;
+
+  if (!PyArg_ParseTuple (args, "O|O", &cb, &cb_context))
+    return NULL;
+
+  if (cb == Py_None && cb_context != NULL) {
+    PyErr_SetString (PyExc_TypeError, "Default callback takes no context");
+    return NULL;
+  }
+  else if (cb != Py_None && !PyCallable_Check (cb)) {
+    PyErr_SetString (PyExc_TypeError, "Parameter must be callable");
+    return NULL;
+  }
+
+  debugprintf ("-> cups_setPasswordCB2\n");
+
+  Py_XINCREF (cb_context);
+  Py_XDECREF (tls->cups_password_callback_context);
+  tls->cups_password_callback_context = cb_context;
+
+  if (cb == Py_None)
+  {
+    Py_XDECREF (tls->cups_password_callback);
+    tls->cups_password_callback = NULL;
+    cupsSetPasswordCB2 (NULL, NULL);
+  }
+  else
+  {
+    Py_XINCREF (cb);
+    Py_XDECREF (tls->cups_password_callback);
+    tls->cups_password_callback = cb;
+    cupsSetPasswordCB2 (password_callback_newstyle, cb_context);
+  }
+
+  debugprintf ("<- cups_setPasswordCB2\n");
+  Py_INCREF (Py_None);
+  return Py_None;
+}
+#endif /* HAVE_CUPS_1_4 */
+
+static PyObject *
+cups_ppdSetConformance (PyObject *self, PyObject *args)
+{
+  int level;
+  if (!PyArg_ParseTuple (args, "i", &level))
+    return NULL;
+
+  ppdSetConformance (level);
   Py_INCREF (Py_None);
   return Py_None;
 }
@@ -339,6 +518,25 @@ static PyMethodDef CupsMethods[] = {
     "@type fn: callable object\n"
     "@param fn: callback function" },
 
+#ifdef HAVE_CUPS_1_4
+  { "setPasswordCB2", cups_setPasswordCB2, METH_VARARGS,
+    "setPasswordCB2(fn, context=None) -> None\n\n"
+    "Set password callback function.  This Python function will be called \n"
+    "when a password is required.  It must take parameters of type string \n"
+    "(the password prompt), instance (the cups.Connection), string (the HTTP "
+    "method), string (the HTTP resource) and, optionally, the user-supplied "
+    "context.  It must return a string (the password).  To \n"
+    "abort the operation it may return the empty string ('').\n\n"
+    "@type fn: callable object, or None for default handler\n"
+    "@param fn: callback function" },
+#endif /* HAVE_CUPS_1_4 */
+
+  { "ppdSetConformance", cups_ppdSetConformance, METH_VARARGS,
+    "ppdSetConformance(level) -> None\n\n"
+    "Set PPD conformance level.\n\n"
+    "@type level: integer\n"
+    "@param level: PPD_CONFORM_RELAXED or PPD_CONFORM_STRICT" },
+
   { "require", cups_require, METH_VARARGS,
     "require(version) -> None\n\n"
     "Require pycups version.\n\n"
@@ -412,6 +610,22 @@ initcups (void)
   PyModule_AddObject (m, "Dest",
 		      (PyObject *)&cups_DestType);
 
+  // IPPRequest type
+  cups_IPPRequestType.tp_new = PyType_GenericNew;
+  if (PyType_Ready (&cups_IPPRequestType) < 0)
+      return;
+
+  PyModule_AddObject (m, "IPPRequest",
+		      (PyObject *)&cups_IPPRequestType);
+
+  // IPPAttribute type
+  cups_IPPAttributeType.tp_new = PyType_GenericNew;
+  if (PyType_Ready (&cups_IPPAttributeType) < 0)
+      return;
+
+  PyModule_AddObject (m, "IPPAttribute",
+		      (PyObject *)&cups_IPPAttributeType);
+
   // Constants
 
 #define INT_CONSTANT(name)					\
@@ -446,6 +660,9 @@ initcups (void)
   INT_CONSTANT (CUPS_PRINTER_AUTHENTICATED);
   INT_CONSTANT (CUPS_PRINTER_COMMANDS);
   INT_CONSTANT (CUPS_PRINTER_OPTIONS);
+#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 3)
+  INT_CONSTANT (CUPS_PRINTER_DISCOVERED);
+#endif /* CUPS 1.3 */
 
   // HTTP encryption
   INT_CONSTANT (HTTP_ENCRYPT_IF_REQUESTED);
@@ -456,6 +673,7 @@ initcups (void)
   // Selected HTTP status codes
   INT_CONSTANT (HTTP_ERROR);
   INT_CONSTANT (HTTP_OK);
+  INT_CONSTANT (HTTP_NOT_MODIFIED);
   INT_CONSTANT (HTTP_BAD_REQUEST);
   INT_CONSTANT (HTTP_UNAUTHORIZED);
   INT_CONSTANT (HTTP_FORBIDDEN);
@@ -469,6 +687,15 @@ initcups (void)
   INT_CONSTANT (PPD_UI_PICKONE);
   INT_CONSTANT (PPD_UI_PICKMANY);
 
+  // PPD Order dependency enum
+
+  INT_CONSTANT (PPD_ORDER_ANY);
+  INT_CONSTANT (PPD_ORDER_DOCUMENT);
+  INT_CONSTANT (PPD_ORDER_EXIT);
+  INT_CONSTANT (PPD_ORDER_JCL);
+  INT_CONSTANT (PPD_ORDER_PAGE);
+  INT_CONSTANT (PPD_ORDER_PROLOG);
+
   // Job states
   INT_CONSTANT (IPP_JOB_PENDING);
   INT_CONSTANT (IPP_JOB_HELD);
@@ -483,6 +710,10 @@ initcups (void)
   INT_CONSTANT (IPP_PRINTER_PROCESSING);
   INT_CONSTANT (IPP_PRINTER_STOPPED);
 
+  // IPP resolution units
+  INT_CONSTANT (IPP_RES_PER_CM);
+  INT_CONSTANT (IPP_RES_PER_INCH);
+
   // IPP finishings
   INT_CONSTANT (IPP_FINISHINGS_NONE);
   INT_CONSTANT (IPP_FINISHINGS_STAPLE);
@@ -513,6 +744,17 @@ initcups (void)
   INT_CONSTANT (IPP_FINISHINGS_BIND_RIGHT);
   INT_CONSTANT (IPP_FINISHINGS_BIND_BOTTOM);
 
+  // IPP orientations
+  INT_CONSTANT (IPP_PORTRAIT);
+  INT_CONSTANT (IPP_LANDSCAPE);
+  INT_CONSTANT (IPP_REVERSE_PORTRAIT);
+  INT_CONSTANT (IPP_REVERSE_LANDSCAPE);
+
+  // IPP qualities
+  INT_CONSTANT (IPP_QUALITY_DRAFT);
+  INT_CONSTANT (IPP_QUALITY_NORMAL);
+  INT_CONSTANT (IPP_QUALITY_HIGH);
+
   // IPP errors
   INT_CONSTANT (IPP_OK);
   INT_CONSTANT (IPP_OK_SUBST);
@@ -559,16 +801,45 @@ initcups (void)
   INT_CONSTANT (IPP_MULTIPLE_JOBS_NOT_SUPPORTED);
   INT_CONSTANT (IPP_PRINTER_IS_DEACTIVATED);
 
+  // IPP states
+  INT_CONSTANT (IPP_ERROR);
+  INT_CONSTANT (IPP_IDLE);
+  INT_CONSTANT (IPP_HEADER);
+  INT_CONSTANT (IPP_ATTRIBUTE);
+  INT_CONSTANT (IPP_DATA);
+
+  // IPP attribute tags
+  INT_CONSTANT (IPP_TAG_ZERO);
+  INT_CONSTANT (IPP_TAG_OPERATION);
+  INT_CONSTANT (IPP_TAG_JOB);
+  INT_CONSTANT (IPP_TAG_PRINTER);
+  INT_CONSTANT (IPP_TAG_INTEGER);
+  INT_CONSTANT (IPP_TAG_BOOLEAN);
+  INT_CONSTANT (IPP_TAG_ENUM);
+  INT_CONSTANT (IPP_TAG_STRING);
+  INT_CONSTANT (IPP_TAG_RANGE);
+  INT_CONSTANT (IPP_TAG_TEXT);
+  INT_CONSTANT (IPP_TAG_NAME);
+  INT_CONSTANT (IPP_TAG_KEYWORD);
+  INT_CONSTANT (IPP_TAG_URI);
+  INT_CONSTANT (IPP_TAG_CHARSET);
+  INT_CONSTANT (IPP_TAG_LANGUAGE);
+  INT_CONSTANT (IPP_TAG_MIMETYPE);
+
   // Limits
   INT_CONSTANT (IPP_MAX_NAME);
 
+  // PPD conformance levels
+  INT_CONSTANT (PPD_CONFORM_RELAXED);
+  INT_CONSTANT (PPD_CONFORM_STRICT);
+
   // Admin Util constants
   STR_CONSTANT (CUPS_SERVER_DEBUG_LOGGING);
   STR_CONSTANT (CUPS_SERVER_REMOTE_ADMIN);
   STR_CONSTANT (CUPS_SERVER_REMOTE_PRINTERS);
   STR_CONSTANT (CUPS_SERVER_SHARE_PRINTERS);
   STR_CONSTANT (CUPS_SERVER_USER_CANCEL_ANY);
-#ifdef HAVE_CUPS_1_3
+#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 3)
   STR_CONSTANT (CUPS_SERVER_REMOTE_ANY);
 #endif /* CUPS 1.3 */
 
